#include "ThermometerApp.h"
#include "stm32g0xx_hal.h"
#include "DynamicTag.h"
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "MAX31875.h"

static MAX31875_Device thermometer;
static MAX31875_Configuration thermometerConfiguration;
static int isThermoMeterInitialized = 0;
static float lastTemperature = -5000;
static uint32_t lastTemperatureTime = -10;
static uint32_t tickCounter = 0;

void SystemClock_Config(void);

extern RTC_HandleTypeDef hrtc;

static void ThermoMeterApp_LowPowerDelay() {
	HAL_StatusTypeDef hStatus;

    HAL_SuspendTick();

    // 8000 = 4s
    hStatus = HAL_RTCEx_SetWakeUpTimer_IT(&hrtc, 8000, RTC_WAKEUPCLOCK_RTCCLK_DIV16);
    if (hStatus) {
    	HAL_Delay(4000);
    }

    HAL_PWR_EnterSTOPMode(PWR_LOWPOWERREGULATOR_ON, PWR_STOPENTRY_WFI);

    HAL_RTCEx_DeactivateWakeUpTimer(&hrtc);
    SystemClock_Config();

    HAL_ResumeTick();

    tickCounter++;
}

static void ThermoMeterApp_InitThermometer() {
	MAX31875_Status mStatus;

	mStatus = MAX31875_Init(&thermometer, MAX31875_I2C_7BIT_ADDRESS_R2);
	if (mStatus) {
		isThermoMeterInitialized = 0;
		return;
	}

	MAX31875_GetDefaultConfiguration(&thermometer, &thermometerConfiguration);
	thermometerConfiguration.conversionRate = MAX31875_ConversionRate_0_25_PerSecond;
	thermometerConfiguration.resolution = MAX31875_Resolution_12bit;
	thermometerConfiguration.dataFormat = MAX31875_DataFormat_Extended;
	thermometerConfiguration.pecMode = MAX31875_PacketErrorCheckingMode_Enabled;

	mStatus = MAX31875_SetConfiguration(&thermometer, &thermometerConfiguration);
	if (mStatus) {
		isThermoMeterInitialized = 0;
		return;
	}

	ThermoMeterApp_LowPowerDelay();

	isThermoMeterInitialized = 1;
	return;
}

int ThermoMeterApp_WriteTag(float temperature) {
	uint8_t tagVal[12 * 4] = {
			0xE1, 0x40, 0x20, 0x01,
			0x03, 0x27, 0xD1, 0x01,
			0x23, 0x54, 0x02, 0x65,
			0x6E, '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', '@', '@', '@',
			'@', 0xFE, 0x00, 0x00 };

	int temperature1000;
	if (temperature > 0) {
		temperature1000 = (int) (temperature * 1000);
	} else {
		temperature1000 = (int) (temperature * -1000);
	}

	char* sign = temperature > 0 ? "" : "-";
	int beforeDot = temperature1000 / 1000;
	int afterDot = temperature1000 % 1000;

	// fill message by spaces
	memset(tagVal + 13, ' ', 32);

	// format message
	snprintf((char*) tagVal + 13, 32, "Temperature: %s%d.%03d °C\n", sign, beforeDot, afterDot);

	// replace \0 terminator by space
	tagVal[13 + strlen((char*) tagVal + 13)] = ' ';

	// write to tag
	return DynamicTag_WriteMem(tagVal, sizeof(tagVal));
}

void ThermoMeterApp_Run() {
	MAX31875_Status mStatus;
	HAL_StatusTypeDef hStatus;
	int iStatus;

	hStatus = HAL_RTCEx_DeactivateWakeUpTimer(&hrtc);
	if (hStatus) {
		__BKPT(0);
	}

	while (1) {
		if (!isThermoMeterInitialized) {
			ThermoMeterApp_InitThermometer();
		}

		if (!isThermoMeterInitialized) {
			ThermoMeterApp_LowPowerDelay();
			continue;
		}

		float temperature;

		mStatus = MAX31875_ReadTemperature(&thermometer, &temperature);
		if (mStatus) {
			isThermoMeterInitialized = 0;
			continue;
		}

		float diff = temperature - lastTemperature;
		if (diff < 0) {
			diff = -diff;
		}

		if (diff > 0.3 && tickCounter - lastTemperatureTime > 5) {
			iStatus = ThermoMeterApp_WriteTag(temperature);
			if (iStatus) {
				ThermoMeterApp_LowPowerDelay();
				continue;
			}

			lastTemperatureTime = tickCounter;
			lastTemperature = temperature;

			HAL_GPIO_WritePin(GPIOB, GPIO_PIN_8, 0);
			HAL_Delay(50);
			HAL_GPIO_WritePin(GPIOB, GPIO_PIN_8, 1);
		}

		ThermoMeterApp_LowPowerDelay();
	}
}
